# pylint: disable=misplaced-comparison-constant,no-self-use

import os
from unittest.mock import patch, Mock

import pytest

from mine import services

from mine.tests.conftest import FILES


class TestFindRoot:

    @patch('os.listdir', Mock(return_value=[]))
    @patch('os.getenv', Mock(return_value=True))
    def test_ci_workaround_enabled(self):
        root = services.find_root(top="mock/top")
        assert "mock/top" == root

    @patch('os.listdir', Mock(return_value=[]))
    @patch('os.getenv', Mock(return_value=False))
    def test_ci_workaround_disabled(self):
        with pytest.raises(EnvironmentError):
            services.find_root(top="mock/top")


class TestFindConfigPath:

    def test_find_dropbox(self, tmpdir):
        """Verify a settings file can be found in Dropbox."""
        tmpdir.chdir()
        _touch('Dropbox', 'mine.yml')

        path = services.find_config_path(tmpdir.strpath)

        assert os.path.isfile(path)

    def test_find_dropbox_personal(self, tmpdir):
        """Verify a settings file can be found in Dropbox (Personal)."""
        tmpdir.chdir()
        _touch('Dropbox (Personal)', 'mine.yml')

        path = services.find_config_path(tmpdir.strpath)

        assert os.path.isfile(path)

    @patch('mine.services.DEPTH', 2)
    def test_find_depth(self, tmpdir):
        """Verify a settings file is not found below the maximum depth."""
        tmpdir.chdir()
        _touch('Dropbox', 'a', 'b', 'mine.yml')

        with pytest.raises(OSError):
            services.find_config_path(tmpdir.strpath)

    def test_find_no_share(self):
        """Verify an error occurs when no service directory is found."""
        with pytest.raises(EnvironmentError):
            with patch('os.getenv', Mock(return_value=False)):
                services.find_config_path(FILES)


@patch('os.remove')
class TestDeleteConflicts:

    @staticmethod
    def _create_conflicts(tmpdir, count=2):
        tmpdir.chdir()
        root = str(tmpdir)

        for index in range(count):
            fmt = "{} (Jace's conflicted copy 2015-03-11).fake"
            filename = fmt.format(index)
            _touch(root, filename)

        return root

    def test_fail_when_leftover_conflicts(self, _, tmpdir):
        root = self._create_conflicts(tmpdir)

        result = services.delete_conflicts(root)

        assert False is result

    def test_pass_when_no_conflicts(self, _, tmpdir):
        root = self._create_conflicts(tmpdir, count=0)

        result = services.delete_conflicts(root)

        assert True is result

    def test_no_deletion_without_force(self, mock_remove, tmpdir):
        root = self._create_conflicts(tmpdir)

        services.delete_conflicts(root, force=False)

        assert 0 == mock_remove.call_count

    def test_deletion_count_is_correct(self, mock_remove, tmpdir):
        root = self._create_conflicts(tmpdir, count=2)

        services.delete_conflicts(root, force=True)

        assert 2 == mock_remove.call_count


def _touch(*parts):
    """Create an empty file at the given path."""
    path = os.path.join(*parts)
    dirpath = os.path.dirname(path)
    if not os.path.isdir(dirpath):
        os.makedirs(dirpath)
    open(path, 'w').close()



from .application import Application, Applications
from .computer import Computer, Computers
from .config import ProgramConfig
from .data import Data
from .status import State, Status, ProgramStatus
from .timestamp import Timestamp



"""Data structures for all settings."""

import logging

import yorm

from .application import Applications
from .computer import Computers


log = logging.getLogger(__name__)


@yorm.attr(computers=Computers)
@yorm.attr(applications=Applications)
class ProgramConfig(yorm.types.AttributeDictionary):
    """Dictionary of program configuration settings."""

    def __init__(self, applications=None, computers=None):
        super().__init__()
        self.applications = applications or Applications()
        self.computers = computers or Computers()



"""Data structures for application information."""

import logging

import yorm

from ._bases import NameMixin


log = logging.getLogger(__name__)


@yorm.attr(mac=yorm.types.NullableString)
@yorm.attr(windows=yorm.types.NullableString)
@yorm.attr(linux=yorm.types.NullableString)
class Versions(yorm.types.AttributeDictionary):
    """Dictionary of OS-specific application filenames."""


@yorm.attr(auto_queue=yorm.types.Boolean)
@yorm.attr(single_instance=yorm.types.Boolean)
class Properties(yorm.types.AttributeDictionary):
    """Dictionary of application management settings."""


@yorm.attr(name=yorm.types.String)
@yorm.attr(properties=Properties)
@yorm.attr(versions=Versions)
class Application(NameMixin, yorm.types.AttributeDictionary):
    """Dictionary of application information."""

    def __init__(self, name, properties=None, filename=None, versions=None):
        super().__init__()
        self.name = name
        self.properties = properties or Properties()
        self.versions = versions or Versions(mac=filename,
                                             windows=filename,
                                             linux=filename)

    @property
    def auto_queue(self):
        return self.properties.auto_queue

    @property
    def no_wait(self):
        return not self.properties.single_instance


@yorm.attr(all=Application)
class Applications(yorm.types.SortedList):
    """List of monitored applications."""

    def get(self, name):
        """Get the application with the given name."""
        application = self.find(name)
        assert application, name
        return application

    def find(self, name):
        """Find the application with the given name, else None."""
        log.debug("Finding application for '%s'...", name)
        for application in self:
            if application == name:
                return application



"""Integration tests configuration file."""

import os

from mine.tests.conftest import pytest_configure, pytest_runtest_setup  # pylint: disable=unused-import

ROOT = os.path.dirname(__file__)
FILES = os.path.join(ROOT, 'files')



"""Classes to manage application state."""

import os
import abc
import time
import glob
import platform
import functools
import subprocess
import logging

import psutil


log = logging.getLogger(__name__)

# TODO: delete this after implementing `BaseManager`
# https://github.com/jacebrowning/mine/issues/8
# https://github.com/jacebrowning/mine/issues/9


# TODO: enable coverage when a Linux test is implemented
def log_running(func):  # pragma: no cover (manual)
    """Decorator for methods that return application status."""
    @functools.wraps(func)
    def wrapped(self, application):
        """Wrapped method to log if an application is running."""
        log.debug("Determining if %s is running...", application)
        running = func(self, application)
        if running is None:
            status = "Untracked"
        elif running:
            status = "Running"
        else:
            status = "Not running"
        log.info("%s: %s", status, application)
        return running
    return wrapped


# TODO: enable coverage when a Linux test is implemented
def log_starting(func):  # pragma: no cover (manual)
    """Decorator for methods that start an application."""
    @functools.wraps(func)
    def wrapped(self, application):
        """Wrapped method to log that an application is being started."""
        log.info("Starting %s...", application)
        result = func(self, application)
        log.info("Running: %s", application)
        return result
    return wrapped


# TODO: enable coverage when a Linux test is implemented
def log_stopping(func):  # pragma: no cover (manual)
    """Decorator for methods that stop an application."""
    @functools.wraps(func)
    def wrapped(self, application):
        """Wrapped method to log that an application is being stopped."""
        log.info("Stopping %s...", application)
        result = func(self, application)
        log.info("Not running: %s", application)
        return result
    return wrapped


class BaseManager(metaclass=abc.ABCMeta):  # pragma: no cover (abstract)
    """Base application manager."""

    NAME = FRIENDLY = None

    IGNORED_APPLICATION_NAMES = []

    def __str__(self):
        return self.FRIENDLY

    @abc.abstractmethod
    def is_running(self, application):
        """Determine if an application is currently running."""
        raise NotImplementedError

    @abc.abstractmethod
    def start(self, application):
        """Start an application on the current computer."""
        raise NotImplementedError

    @abc.abstractmethod
    def stop(self, application):
        """Stop an application on the current computer."""
        raise NotImplementedError

    @abc.abstractmethod
    def launch(self, path):
        """Open a file for editing."""
        raise NotImplementedError

    @classmethod
    def _get_process(cls, name):
        """Get a process whose executable path contains an app name."""
        log.debug("Searching for exe path containing '%s'...", name)

        for process in psutil.process_iter():
            try:
                command = ' '.join(process.cmdline()).lower()
                parts = []
                for arg in process.cmdline():
                    parts.extend([p.lower() for p in arg.split(os.sep)])
            except psutil.AccessDenied:
                continue  # the process is likely owned by root

            if name.lower() not in parts:
                continue

            if process.pid == os.getpid():
                log.debug("Skipped current process: %s", command)
                continue

            if process.status() == psutil.STATUS_ZOMBIE:
                log.debug("Skipped zombie process: %s", command)
                continue

            log.debug("Found matching process: %s", command)
            for ignored in cls.IGNORED_APPLICATION_NAMES:
                if ignored.lower() in parts:
                    log.debug("But skipped due to ignored name")
                    break
            else:
                return process

        return None


class LinuxManager(BaseManager):  # pragma: no cover (manual)
    """Application manager for Linux."""

    NAME = 'Linux'
    FRIENDLY = NAME

    def is_running(self, application):
        name = application.versions.linux
        if not name:
            return None
        process = self._get_process(name)
        return process is not None

    def start(self, application):
        pass

    def stop(self, application):
        name = application.versions.linux
        process = self._get_process(name)
        if process.is_running():
            process.terminate()

    def launch(self, path):
        log.info("Opening %s...", path)
        return subprocess.call(['xdg-open', path]) == 0


class MacManager(BaseManager):  # pragma: no cover (manual)
    """Application manager for OS X."""

    NAME = 'Darwin'
    FRIENDLY = 'Mac'

    IGNORED_APPLICATION_NAMES = [
        "iTunesHelper.app",
    ]

    @log_running
    def is_running(self, application):
        name = application.versions.mac
        if not name:
            return None
        process = self._get_process(name)
        return process is not None

    @log_starting
    def start(self, application):
        name = application.versions.mac
        path = None
        for base in (".",
                     "~/Applications",
                     "/Applications",
                     "/Applications/*"):
            pattern = os.path.expanduser(os.path.join(base, name))
            log.debug("Glob pattern: %s", pattern)
            paths = glob.glob(pattern)
            if paths:
                path = paths[0]
                log.debug("Match: %s", path)
                break
        else:
            assert path, "Not found: {}".format(application)
        return self._start_app(path)

    @log_stopping
    def stop(self, application):
        name = application.versions.mac
        process = self._get_process(name)
        if process and process.is_running():
            process.terminate()

    @staticmethod
    def _start_app(path):
        """Start an application from it's .app directory."""
        assert os.path.exists(path), path
        process = psutil.Popen(['open', path])
        time.sleep(1)
        return process

    def launch(self, path):
        log.info("opening %s...", path)
        return subprocess.call(['open', path]) == 0


class WindowsManager(BaseManager):  # pragma: no cover (manual)
    """Application manager for Windows."""

    NAME = 'Windows'
    FRIENDLY = NAME

    def is_running(self, application):
        pass

    def start(self, application):
        pass

    def stop(self, application):
        pass

    def launch(self, path):
        log.info("starting %s...", path)
        os.startfile(path)  # pylint: disable=no-member
        return True


def get_manager(name=None):
    """Return an application manager for the current operating system."""
    log.info("Detecting the current system...")
    name = name or platform.system()
    manager = {
        WindowsManager.NAME: WindowsManager,
        MacManager.NAME: MacManager,
        LinuxManager.NAME: LinuxManager,
    }[name]()
    log.info("Current system: %s", manager)
    return manager



# pylint: disable=misplaced-comparison-constant,no-self-use

import pytest

from mine.models import Application, Applications


class TestApplication:
    """Unit tests for the application class."""

    app1 = Application('iTunes')
    app1.versions.mac = 'iTunes.app'
    app2 = Application('HipChat')
    app3 = Application('Sublime Text')
    app3.versions.linux = 'sublime_text'
    app4 = Application('hipchat')

    str_application = [
        ("iTunes", app1),
        ("HipChat", app2),
        ("Sublime Text", app3),
        ("hipchat", app4),
    ]

    @pytest.mark.parametrize("string,application", str_application)
    def test_str(self, string, application):
        """Verify applications can be converted to strings."""
        assert string == str(application)

    def test_eq(self):
        """Verify applications can be equated."""
        assert self.app2 == self.app4
        assert self.app1 != self.app3

    def test_lt(self):
        """Verify applications can be sorted."""
        assert self.app2 < self.app1
        assert self.app3 > self.app2


class TestApplications:
    """Unit tests for lists of applications."""

    apps = Applications([TestApplication.app1, TestApplication.app2])

    def test_get(self):
        """Verify an application can be found in a list."""
        app = self.apps.get('itunes')
        assert 'iTunes' == app.name

    def test_get_missing(self):
        """Verify an invalid names raise an assertion."""
        with pytest.raises(AssertionError):
            self.apps.get('fake')



"""Data structures for computer information."""

import uuid
import socket
import logging

import yorm

from ._bases import NameMixin


log = logging.getLogger(__name__)


@yorm.attr(name=yorm.types.String)
@yorm.attr(hostname=yorm.types.String)
@yorm.attr(address=yorm.types.String)
class Computer(NameMixin, yorm.types.AttributeDictionary):
    """A dictionary of identifying computer information."""

    def __init__(self, name, hostname=None, address=None):
        super().__init__()
        self.name = name
        self.address = address or self.get_address()
        self.hostname = hostname or self.get_hostname()

    @staticmethod
    def get_address(node=None):
        """Get this computer's MAC address."""
        if node is None:
            node = uuid.getnode()
        return ':'.join(("%012X" % node)[i:i + 2] for i in range(0, 12, 2))

    @staticmethod
    def get_hostname():
        """Get this computer's hostname."""
        return socket.gethostname()


@yorm.attr(all=Computer)
class Computers(yorm.types.SortedList):
    """A list of computers."""

    @property
    def names(self):
        """Get a list of all computers' labels."""
        return [c.name for c in self]

    def get(self, name):
        """Get the computer with the given name."""
        computer = self.find(name)
        assert computer, name
        return computer

    def find(self, name):
        """Find the computer with the given name, else None."""
        log.debug("Finding computer for '%s'...", name)
        for computer in self:
            if computer == name:
                return computer

    def match(self, partial):
        """Find a computer with a similar name."""
        log.debug("Finding computer similar to '%s'...", partial)
        for computer in self:
            if partial.lower() in computer.name.lower():
                return computer

    def get_current(self):
        """Get the current computer's information."""
        this = Computer(None)

        # Search for a matching address
        for other in self:
            if this.address == other.address:
                other.hostname = this.hostname
                return other

        # Else, search for a matching hostname
        for other in self:
            if this.hostname == other.hostname:
                other.address = this.address
                return other

        # Else, this is a new computer
        this.name = self.generate_name(this)
        assert this.name != 'localhost'
        log.debug("New computer: %s", this)
        self.append(this)
        return this

    def generate_name(self, computer):
        """Generate a new label for a computer."""
        name = computer.hostname.lower().split('.')[0]
        copy = 1
        while name in self.names:
            copy += 1
            name2 = "{}-{}".format(name, copy)
            if name2 not in self.names:
                name = name2
        return name



# pylint: disable=misplaced-comparison-constant,no-self-use

import os
import subprocess
import logging

import yorm

from mine import cli
from mine.models import (Application, Computer, Data, ProgramConfig,
                         State, Status, ProgramStatus)

from .conftest import FILES


log = logging.getLogger(__name__)


class TestFiles:
    """Integration tests for creating files."""

    def test_data(self):
        """Verify a sample file is created."""
        path = os.path.join(FILES, 'mine.yml')

        if os.path.exists(path):
            os.remove(path)

        data = Data()
        yorm.sync(data, path)

        itunes = Application('itunes')
        itunes.versions.mac = ''
        itunes.versions.windows = 'iTunes.exe'

        iphoto = Application('iphoto')
        iphoto.versions.mac = 'iPhoto'

        mac = Computer('macbook', 'Other.local')
        mac2 = Computer('macbook-pro')

        configuration = ProgramConfig()
        configuration.applications = [itunes, iphoto]
        configuration.computers = [mac, mac2]

        data.config = configuration

        mac_state = State('macbook-pro')
        mac_state.timestamp.started = 444

        itunes_status = Status('itunes')
        itunes_status.computers = [mac_state]

        status = ProgramStatus()
        status.applications = [itunes_status]
        status.counter = 499

        data.status = status

        assert os.path.exists(path)

    def test_data_out(self):
        """Verify a sample file is created."""
        path = os.path.join(FILES, 'mine-out.yml')

        if os.path.exists(path):
            os.remove(path)

        data = Data()
        yorm.sync(data, path)

        itunes = Application('itunes')
        itunes.versions.mac = ''
        itunes.versions.windows = 'iTunes.exe'

        iphoto = Application('iphoto')
        iphoto.versions.mac = 'iPhoto'

        mac = Computer('macbook', 'Jaces-MacBook', 'AA:BB:CC:DD:EE:FF')
        mac2 = Computer('macbook-pro', 'Jaces-MacBook-2', '11:22:33:44:55:66')

        configuration = ProgramConfig()
        configuration.applications = [itunes, iphoto]
        configuration.computers = [mac, mac2]

        data.config = configuration

        mac_state = State('macbook-pro')
        mac_state.timestamp.started = 444

        itunes_status = Status('itunes')
        itunes_status.computers = [mac_state]

        status = ProgramStatus()
        status.applications = [itunes_status]
        status.counter = 499

        data.status = status

        assert os.path.exists(path)

    def test_data_in(self):
        """Verify a sample file is loaded."""
        path = os.path.join(FILES, 'mine-in.yml')

        data = Data()
        yorm.sync(data, path)

        assert data.config.applications
        for application in data.config.applications:
            if application.name == 'slack':
                break
        else:
            assert False


class TestProcesses:
    """Integration tests for tracking and stopping processes."""

    NAME = "example application"

    application = Application(NAME)
    application.versions.linux = 'yes'
    application.versions.mac = 'yes'
    application.versions.windows = 'yes.exe'

    computer = None
    data = None

    path = os.path.join(FILES, 'mine.tmp.yml')

    _process = None

    def _store_data(self):
        """Set up initial data file for tests."""
        self.data = Data()
        self.data.config.applications.append(self.application)
        self.computer = self.data.config.computers.get_current()  # pylint: disable=no-member
        yorm.sync(self.data, self.path)

    def _fetch_data(self):
        """Read the final data file back for verification."""
        data = Data()
        yorm.sync(data, self.path)
        return data

    def _start_application(self):
        """Start the example application."""
        if not self._process:
            # TODO: get filename from the application object
            self._process = subprocess.Popen(['yes'], stdout=subprocess.PIPE)
        log.info("%s is started", self.application)

    def _stop_application(self):
        """Stop the example application."""
        if self._process:
            if self._process.poll() is None:
                self._process.kill()
            self._process = None
        log.info("%s is stopped", self.application)

    def _is_application_running(self):
        """Determine if the sample application is running."""
        return self._process and self._process.poll() is None

    def teardown_method(self, _):
        """Stop the sample application and clean up the file."""
        self._stop_application()
        if os.path.exists(self.path):
            os.remove(self.path)

    def test_case_1(self):
        """Verify a newly running remote application takes over."""

        # Arrange

        self._store_data()

        # start the application and manually mark it as running
        self._start_application()
        status = self.data.status
        status.start(self.application, self.computer)
        self.data.status = status
        assert self.data.status.is_running(self.application, self.computer)

        # manually mark the application as running on a remote computer
        computer = Computer('other', 'Other.local', 'AA:BB:CC:DD:EE:FF')
        status = self.data.status
        status.start(self.application, computer)
        self.data.status = status
        assert self.data.status.is_running(self.application, computer)

        # Act

        cli.run(self.path, cleanup=False)

        # Assert

        # verify the application is no longer running
        assert not self._is_application_running()
        # verify the application is marked as running remotely
        data = self._fetch_data()
        assert not data.status.is_running(self.application, self.computer)
        assert data.status.is_running(self.application, computer)

    def test_case_2(self):
        """Verify a newly running local application takes over."""

        # Arrange

        self._store_data()

        # manually mark the application as running on a remote computer
        computer = Computer('other', 'Other.local', 'AA:BB:CC:DD:EE:FF')
        status = self.data.status
        status.start(self.application, computer)
        self.data.status = status
        assert self.data.status.is_running(self.application, computer)

        # start the application
        self._start_application()

        # Act
        cli.run(self.path, cleanup=False)

        # Assert

        assert self._is_application_running()

        data = self._fetch_data()
        assert data.status.is_running(self.application, self.computer)
        assert data.status.is_running(self.application, computer)

    def test_case_3(self):
        """Verify an already running local application is ignored."""

        # Arrange

        self._store_data()

        status = self.data.status
        status.start(self.application, self.computer)
        self.data.status = status

        self._start_application()

        # Act

        cli.run(self.path)

        # Assert

        assert self._is_application_running()

        data = self._fetch_data()
        assert data.status.is_running(self.application, self.computer)

    def test_case_4(self):
        """Verify a newly stopped local application is noted."""

        # Arrange

        self._store_data()

        status = self.data.status
        status.start(self.application, self.computer)
        self.data.status = status

        self._stop_application()

        # Act

        cli.run(self.path)

        # Assert

        assert not self._is_application_running()

        data = self._fetch_data()
        assert not data.status.is_running(self.application, self.computer)

    def test_case_5(self):
        """Verify an already stopped local application is ignored."""

        # Arrange

        self._store_data()

        self._stop_application()

        # Act

        cli.run(self.path)

        # Assert

        assert not self._is_application_running()

        data = self._fetch_data()
        assert not data.status.is_running(self.application, self.computer)



"""Data structures for timestamp information."""

import logging

import yorm


log = logging.getLogger(__name__)


@yorm.attr(started=yorm.types.Integer)
@yorm.attr(stopped=yorm.types.Integer)
class Timestamp(yorm.types.AttributeDictionary):
    """Dictionary of last start and stop times."""

    def __init__(self, started=0, stopped=0):
        super().__init__()
        self.started = started
        self.stopped = stopped

    def __repr__(self):
        return "<timestamp {}>".format(self.latest)

    def __eq__(self, other):
        return self.latest == other.latest

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return self.latest < other.latest

    @property
    def latest(self):
        """Get the latest timestamp."""
        return max((self.started, self.stopped))

    @property
    def active(self):
        """Determine if the timestamps indicate current activity."""
        if not self.started:
            return False

        if not self.stopped:
            return True

        return self.started > self.stopped



# pylint: disable=misplaced-comparison-constant,unused-variable

import pytest

from mine.models import Data


def describe_data():

    @pytest.fixture
    def data():
        return Data()

    def describe_repr():

        def it_should_always_be_a_simple_name(data):
            assert "settings" == repr(data)

    def describe_modified():

        def is_false_initially(data):
            assert False is data.modified

        def is_true_when_the_counter_changes(data):
            data.status.counter += 1

            assert True is data.modified

        def is_false_after_reading(data):
            data.status.counter += 1

            print(data.modified)

            assert False is data.modified



"""Common exceptions, classes, and functions."""

import argparse
import logging


from . import settings

MAX_VERBOSITY = 4

verbosity = 0


class HelpFormatter(argparse.HelpFormatter):
    """Command-line help text formatter with wider help text."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, max_help_position=40, **kwargs)


class WarningFormatter(logging.Formatter):
    """Logging formatter that displays verbose formatting for WARNING+."""

    def __init__(self, default_format, verbose_format, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_format = default_format
        self.verbose_format = verbose_format

    def format(self, record):
        """Python 3 hack to change the formatting style dynamically."""
        # pylint: disable=protected-access
        if record.levelno > logging.INFO:
            self._style._fmt = self.verbose_format
        else:
            self._style._fmt = self.default_format
        return super().format(record)


def configure_logging(count=0):
    """Configure logging using the provided verbosity count."""
    assert MAX_VERBOSITY == 4

    if count == -1:
        level = settings.QUIET_LOGGING_LEVEL
        default_format = settings.DEFAULT_LOGGING_FORMAT
        verbose_format = settings.LEVELED_LOGGING_FORMAT
    elif count == 0:
        level = settings.DEFAULT_LOGGING_LEVEL
        default_format = settings.DEFAULT_LOGGING_FORMAT
        verbose_format = settings.LEVELED_LOGGING_FORMAT
    elif count == 1:
        level = settings.VERBOSE_LOGGING_LEVEL
        default_format = settings.VERBOSE_LOGGING_FORMAT
        verbose_format = settings.VERBOSE_LOGGING_FORMAT
    elif count == 2:
        level = settings.VERBOSE2_LOGGING_LEVEL
        default_format = settings.VERBOSE_LOGGING_FORMAT
        verbose_format = settings.VERBOSE_LOGGING_FORMAT
    elif count == 3:
        level = settings.VERBOSE2_LOGGING_LEVEL
        default_format = settings.VERBOSE2_LOGGING_FORMAT
        verbose_format = settings.VERBOSE2_LOGGING_FORMAT
    else:
        level = settings.VERBOSE2_LOGGING_LEVEL - 1
        default_format = settings.VERBOSE2_LOGGING_FORMAT
        verbose_format = settings.VERBOSE2_LOGGING_FORMAT

    # Set a custom formatter
    logging.basicConfig(level=level)
    logging.captureWarnings(True)
    formatter = WarningFormatter(default_format, verbose_format,
                                 datefmt=settings.LOGGING_DATEFMT)
    logging.root.handlers[0].setFormatter(formatter)
    logging.getLogger('yorm').setLevel(max(level, settings.YORM_LOGGING_LEVEL))

    # Warn about excessive verbosity
    global verbosity
    if count > MAX_VERBOSITY:
        msg = "Maximum verbosity level is {}".format(MAX_VERBOSITY)
        logging.warning(msg)
        verbosity = MAX_VERBOSITY
    else:
        verbosity = count



# pylint: disable=misplaced-comparison-constant,no-self-use

import os
from unittest.mock import Mock, patch
import logging

import pytest

from mine import cli
from mine import common
from mine.models import Application


class TestMain:
    """Unit tests for the `main` function."""

    @patch('mine.cli.run', Mock(return_value=True))
    def test_run(self):
        """Verify the CLI can be run."""
        cli.main([])

    @patch('mine.cli.run', Mock(return_value=False))
    def test_run_fail(self):
        """Verify the CLI can detect errors."""
        with pytest.raises(SystemExit):
            cli.main([])

    def test_help(self):
        """Verify the CLI help text can be displayed."""
        with pytest.raises(SystemExit):
            cli.main(['--help'])

    @patch('mine.cli.run', Mock(side_effect=KeyboardInterrupt))
    def test_interrupt(self):
        """Verify the CLI can be interrupted."""
        with pytest.raises(SystemExit):
            cli.main([])

    @patch('mine.cli.log')
    @patch('mine.cli.run', Mock(side_effect=KeyboardInterrupt))
    def test_interrupt_verbose(self, mock_log):
        """Verify the CLI can be interrupted (verbose output)."""
        with pytest.raises(SystemExit):
            cli.main(['-vvvv'])
        assert mock_log.exception.call_count == 1

    @patch('mine.cli.daemon', None)
    def test_path(self, path):
        """Verify a custom setting file path can be used."""
        cli.main(['--file', path])

        assert os.path.isfile(path)

    @patch('mine.cli.run')
    def test_daemon(self, mock_run):
        cli.main(['--daemon'])
        mock_run.assert_called_once_with(path=None, delay=300)

    @patch('mine.cli.run')
    def test_daemon_with_specific_delay(self, mock_run):
        cli.main(['--daemon', '42'])
        mock_run.assert_called_once_with(path=None, delay=42)

    @patch('mine.cli.daemon', Application(None))
    def test_warning_when_daemon_is_not_running(self, path):
        with pytest.raises(SystemExit):
            cli.main(['--file', path])


class TestSwitch:
    """Unit tests for the `switch` function."""

    @patch('mine.cli.run')
    def test_switch(self, mock_run):
        """Verify a the current computer can be queued."""
        cli.main(['switch'])
        mock_run.assert_called_once_with(path=None, delay=None,
                                         switch=True)

    @patch('mine.cli.run')
    def test_switch_specific(self, mock_run):
        """Verify a specific computer can be queued."""
        cli.main(['switch', 'foobar'])
        mock_run.assert_called_once_with(path=None, delay=None,
                                         switch='foobar')


class TestClean:

    @patch('mine.cli.run')
    def test_clean(self, mock_run):
        cli.main(['clean'])
        mock_run.assert_called_once_with(path=None, delay=None,
                                         delete=True, force=False)

    @patch('mine.cli.run')
    def test_clean_with_force(self, mock_run):
        cli.main(['clean', '--force'])
        mock_run.assert_called_once_with(path=None, delay=None,
                                         delete=True, force=True)


class TestEdit:

    @patch('mine.cli.run')
    def test_edit(self, mock_run):
        cli.main(['edit'])
        mock_run.assert_called_once_with(path=None, delay=None,
                                         edit=True,)


def _mock_run(*args, **kwargs):
    """Placeholder logic for logging tests."""
    logging.debug(args)
    logging.debug(kwargs)
    logging.warning("warning")
    logging.error("error")
    return True


@patch('mine.cli.run', _mock_run)
class TestLogging:
    """Integration tests for the Doorstop CLI logging."""

    arg_verbosity = [
        ('', 0),
        ('-v', 1),
        ('-vv', 2),
        ('-vvv', 3),
        ('-vvvv', 4),
        ('-vvvvv', 4),
        ('-q', -1),
    ]

    @pytest.mark.parametrize("arg,verbosity", arg_verbosity)
    def test_verbose(self, arg, verbosity):
        """Verify verbose level can be set."""
        cli.main([arg] if arg else [])
        assert verbosity == common.verbosity



"""Unit test configuration file."""

import os
import platform
import logging

import pytest

ENV = 'TEST_INTEGRATION'  # environment variable to enable integration tests
REASON = "'{0}' variable not set".format(ENV)

ROOT = os.path.dirname(__file__)
FILES = os.path.join(ROOT, 'files')


def pytest_configure(config):
    """Disable verbose output when running tests."""
    logging.basicConfig(
        level=logging.DEBUG,
        format="[%(levelname)-8s] (%(name)s @%(lineno)4d) %(message)s",
    )
    logging.getLogger('yorm').setLevel(logging.WARNING)

    terminal = config.pluginmanager.getplugin('terminal')
    base = terminal.TerminalReporter

    class QuietReporter(base):
        """A py.test reporting that only shows dots when running tests."""

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.verbosity = 0
            self.showlongtestinfo = False
            self.showfspath = False

    terminal.TerminalReporter = QuietReporter


def pytest_runtest_setup(item):
    if 'linux_only' in item.keywords and platform.system() != 'Linux':
        pytest.skip("test can only be run on Linux")
    if 'mac_only' in item.keywords and platform.system() != 'Darwin':
        pytest.skip("test can only be run on OS X")
    if 'windows_only' in item.keywords and platform.system() != 'Windows':
        pytest.skip("test can only be run on Windows")


@pytest.fixture
def path(tmpdir):
    """Figure to create a temporary settings file path."""
    tmpdir.chdir()
    return tmpdir.join('custom.ext').strpath



"""Program defaults."""

import logging


# Logging settings
DEFAULT_LOGGING_FORMAT = "%(message)s"
LEVELED_LOGGING_FORMAT = "%(levelname)s: %(message)s"
VERBOSE_LOGGING_FORMAT = "[%(asctime)s %(levelname)8s] %(message)s"
VERBOSE2_LOGGING_FORMAT = "[%(asctime)s %(levelname)8s] (%(name)-13s @%(lineno)4d) %(message)s"  # pylint: disable=C0301
QUIET_LOGGING_LEVEL = logging.WARNING
DEFAULT_LOGGING_LEVEL = logging.WARNING
VERBOSE_LOGGING_LEVEL = logging.INFO
VERBOSE2_LOGGING_LEVEL = logging.DEBUG
LOGGING_DATEFMT = "%Y-%m-%d %H:%M"

# 3rd party settings
YORM_LOGGING_LEVEL = logging.WARNING



# pylint: disable=misplaced-comparison-constant,no-self-use

import pytest

from mine.models import Timestamp


class TestTimestamp:
    """Unit tests for the timestamp class."""

    none = Timestamp()
    started_only = Timestamp(started=42)
    stopped_only = Timestamp(stopped=42)
    started_after_stopped = Timestamp(2, 1)
    stopped_after_started = Timestamp(3, 4)

    active_timestamp = [
        (False, none),
        (True, started_only),
        (False, stopped_only),
        (True, started_after_stopped),
        (False, stopped_after_started),
    ]

    repr_timestamp = [
        ("<timestamp 0>", none),
        ("<timestamp 42>", started_only),
        ("<timestamp 42>", stopped_only),
        ("<timestamp 2>", started_after_stopped),
        ("<timestamp 4>", stopped_after_started),
    ]

    @pytest.mark.parametrize("representation,timestamp", repr_timestamp)
    def test_repr(self, representation, timestamp):
        """Verify timestamps can represented as strings."""
        assert representation == repr(timestamp)

    def test_eq(self):
        """Verify timestamps can be equated."""
        assert Timestamp(1, 1) == Timestamp(1, 1)
        assert Timestamp(1, 2) == Timestamp(1, 2)
        assert Timestamp(1, 2) != Timestamp(1, 3)
        assert Timestamp(4, 2) == Timestamp(4, 3)

    def test_lt(self):
        """Verify timestamps can be sorted."""
        assert Timestamp(1, 2) < Timestamp(1, 3)
        assert Timestamp(1, 4) > Timestamp(1, 3)
        assert Timestamp(4, 2) < Timestamp(5, 3)
        assert Timestamp(3, 2) < Timestamp(4, 3)

    @pytest.mark.parametrize("active,timestamp", active_timestamp)
    def test_active(self, active, timestamp):
        """Verify started/stopped counters determine activity."""
        assert active == timestamp.active



# pylint: disable=misplaced-comparison-constant,no-self-use

import pytest

from mine.manager import get_manager, LinuxManager, MacManager, WindowsManager
from mine.models import Application


class TestLinuxManager:
    """Unit tests for the Linux manager class."""

    manager = get_manager('Linux')

    def test_init(self):
        """Verify the OS is detected correctly."""
        assert isinstance(self.manager, LinuxManager)

    def test_is_running_false(self):
        """Verify a process can be detected as not running."""
        pytest.skip("TODO: implement test")
        application = Application('fake-program-for-mine')
        application.versions.linux = 'fake_program_for_mine.sh'
        assert False is self.manager.is_running(application)

    @pytest.mark.linux_only
    def test_is_running_true(self):
        """Verify a process can be detected as running."""
        pytest.skip("TODO: implement test")

    def test_is_running_none(self):
        """Verify a process can be detected as untracked."""
        application = Application('fake-program-for-mine')
        assert None is self.manager.is_running(application)


class TestMacManager:
    """Unit tests for the Mac manager class."""

    manager = get_manager('Darwin')

    def test_init(self):
        """Verify the OS is detected correctly."""
        assert isinstance(self.manager, MacManager)

    def test_is_running_false(self):
        """Verify a process can be detected as not running."""
        application = Application('fake-program-for-mine')
        application.versions.mac = 'FakeProgramForMine.app'
        assert False is self.manager.is_running(application)

    @pytest.mark.mac_only
    def test_is_running_true(self):
        """Verify a process can be detected as running."""
        application = Application('finder')
        application.versions.mac = 'Finder.app'
        assert True is self.manager.is_running(application)

    def test_is_running_none(self):
        """Verify a process can be detected as untracked."""
        application = Application('fake-program-for-mine')
        assert None is self.manager.is_running(application)

    @pytest.mark.mac_only
    def test_stop(self):
        """Verify a process can be stopped."""
        application = Application('mail')
        application.versions.mac = 'Mail.app'
        self.manager.stop(application)
        assert not self.manager.is_running(application)


class TestWindowsManager:
    """Unit tests for the Windows manager class."""

    manager = get_manager('Windows')

    def test_init(self):
        """Verify the OS is detected correctly."""
        assert isinstance(self.manager, WindowsManager)

    def test_is_running_false(self):
        """Verify a process can be detected as not running."""
        pytest.skip("TODO: implement test")
        application = Application('fake-program-for-mine')
        application.versions.windows = 'FakeProgramForMine.exe'
        assert False is self.manager.is_running(application)

    @pytest.mark.windows_only
    def test_is_running_true(self):
        """Verify a process can be detected as running."""
        pytest.skip("TODO: implement test")
        application = Application('explorer')
        application.versions.windows = 'explorer.exe'
        assert True is self.manager.is_running(application)

    def test_is_running_none(self):
        """Verify a process can be detected as untracked."""
        application = Application('fake-program-for-mine')
        assert None is self.manager.is_running(application)



# pylint: disable=misplaced-comparison-constant,no-self-use

from unittest.mock import Mock, patch

import pytest

from mine.models import Computer, Computers


@patch('uuid.getnode', Mock(return_value=0))
@patch('socket.gethostname', Mock(return_value='Sample.local'))
class TestComputer:
    """Unit tests for the computer classes."""

    def test_init(self):
        """Verify the correct computer information is loaded."""
        computer = Computer('my-sample')
        assert 'my-sample' == computer.name
        assert '00:00:00:00:00:00' == computer.address
        assert 'Sample.local' == computer.hostname

    def test_init_defaults(self):
        """Verify the correct computer information can be overridden."""
        computer = Computer('name', 'hostname', 'address')
        assert 'name' == computer.name
        assert 'address' == computer.address
        assert 'hostname' == computer.hostname

    def test_eq(self):
        """Verify computers and strings can be equated."""
        assert Computer('mac1') == Computer('mac1')
        assert Computer('mac1') != Computer('mac2')
        assert Computer('mac1') == 'mac1'
        assert 'mac1' != Computer('mac2')

    def test_lt(self):
        """Verify computers can be sorted."""
        assert Computer('mac1') < Computer('mac2')
        assert Computer('def') > Computer('ABC')

    def test_get_match_none(self):
        """Verify a computer is added when missing."""
        other = Computer('name', 'hostname', 'address')
        computers = Computers([other])
        this = computers.get_current()
        assert 'sample' == this.name
        assert '00:00:00:00:00:00' == this.address
        assert 'Sample.local' == this.hostname
        assert 2 == len(computers)

    def test_get_match_all(self):
        """Verify a computer can be matched exactly."""
        other = Computer('all', 'Sample.local', '00:00:00:00:00:00')
        computers = Computers([other])
        this = computers.get_current()
        assert 'all' == this.name
        assert '00:00:00:00:00:00' == this.address
        assert 'Sample.local' == this.hostname
        assert 1 == len(computers)


@patch('uuid.getnode', Mock(return_value=0))
@patch('socket.gethostname', Mock(return_value='Sample.local'))
class TestComputers:
    """Unit tests for lists of computers."""

    computers = Computers([Computer('abc', 'abc.local', 1),
                           Computer('def', 'def.local', 2)])

    def test_get(self):
        """Verify a computer can be found in a list."""
        computer = self.computers.get('ABC')
        assert 'abc' == computer.name

    def test_get_missing(self):
        """Verify an invalid names raise an assertion."""
        with pytest.raises(AssertionError):
            self.computers.get('fake')

    def test_match(self):
        """Verify a similar computer can be found."""
        computer = self.computers.match('AB')
        assert 'abc' == computer.name

    def test_generate_name(self):
        """Verify a computer name is generated correctly."""
        computers = Computers()
        computer = Computer(None, hostname='Jaces-iMac.local')
        name = computers.generate_name(computer)
        assert 'jaces-imac' == name

    def test_generate_name_duplicates(self):
        """Verify a computer name is generated correctly with duplicates."""
        computers = Computers([Computer('jaces-imac')])
        computer = Computer(None, hostname='Jaces-iMac.local')
        name = computers.generate_name(computer)
        assert 'jaces-imac-2' == name

    def test_get_current_by_matching_address(self):
        computers = Computers([Computer('abc', 'foobar', '00:00:00:00:00:00')])
        computer = computers.get_current()
        assert 'abc' == computer.name

    def test_get_current_by_matching_hostname(self):
        computers = Computers([Computer('abc', 'Sample.local', 'foobar')])
        computer = computers.get_current()
        assert 'abc' == computer.name



"""Functions to interact with sharing services."""

import os
import re
import logging

from .models.application import Application, Versions


ROOTS = (
    r"C:\Users",
    r"/Users",
    r"/home",
)
SERVICES = (
    'Dropbox',
    'Dropbox (Personal)',
)
CONFIG = 'mine.yml'
CONFLICT_BASE = r"{} \(.+'s conflicted copy \d+-\d+-\d+.*\).*"
CONFLICT_ANY = CONFLICT_BASE.format(".+")
CONFLICT_CONFIG = CONFLICT_BASE.format("mine")
DEPTH = 3  # number of levels to search for the settings file
APPLICATION = Application("Dropbox", versions=Versions(
    mac="Dropbox.app",
    windows="Dropbox.exe",
    linux="dropbox"))

log = logging.getLogger(__name__)


def find_root(top=None):
    """Get the root of the shared directory."""
    top = top or os.path.expanduser('~')

    log.debug("Looking for sharing service in '%s'...", top)
    for directory in os.listdir(top):
        if directory in SERVICES:
            path = os.path.join(top, directory)
            log.debug("Found sharing service: %s", path)
            return path

    msg = "no sharing service found"
    if os.getenv('CI'):
        log.warning(msg)
        return top
    else:
        raise EnvironmentError(msg)


def find_config_path(top=None, root=None):
    """Get the path to the settings file."""
    log.info("Looking for settings file...")
    top = top or os.path.expanduser('~')
    root = root or find_root(top=top)

    log.debug("Looking for '%s' in '%s'...", CONFIG, root)
    for dirpath, dirnames, _, in os.walk(root):
        depth = dirpath.count(os.path.sep) - root.count(os.path.sep)
        if depth >= DEPTH:
            del dirnames[:]
            continue
        path = os.path.join(dirpath, CONFIG)
        if os.path.isfile(path) and \
                not os.path.isfile(os.path.join(path, 'setup.py')):
            log.info("Found settings file: %s", path)
            return path

    raise EnvironmentError("No '{}' file found".format(CONFIG))


def delete_conflicts(root=None, config_only=False, force=False):
    """Delete all files with conflicted filenames."""
    root = root or find_root()

    log.info("%s conflicted files...", 'Deleting' if force else 'Displaying')
    pattern = CONFLICT_CONFIG if config_only else CONFLICT_ANY
    log.debug("Pattern: %r", pattern)
    regex = re.compile(pattern)
    count = 0
    for dirname, _, filenames in os.walk(root):
        for filename in filenames:
            if regex.match(filename):
                count += 1
                path = os.path.join(dirname, filename)
                if force:
                    os.remove(path)
                print(path)

    if count and not force:
        print()
        print("Run again with '--force' to delete these "
              "{} conflict(s)".format(count))
        return False

    return True



"""Package for mine."""

__project__ = 'mine'
__version__ = '1.3'

CLI = 'mine'
VERSION = '{0} v{1}'.format(__project__, __version__)
DESCRIPTION = 'Share application state across computers using Dropbox.'



"""Data structures that combine all program data."""

import logging

import yorm

from .config import ProgramConfig
from .status import ProgramStatus


log = logging.getLogger(__name__)


@yorm.attr(config=ProgramConfig)
@yorm.attr(status=ProgramStatus)
class Data:
    """Primary wrapper for all settings."""

    def __init__(self):
        self.config = ProgramConfig()
        self.status = ProgramStatus()
        self._last_counter = self.status.counter

    def __repr__(self):
        return "settings"

    @property
    def modified(self):
        changed = self.status.counter != self._last_counter
        self._last_counter = self.status.counter
        return changed

    @staticmethod
    def prune_status(config, status):
        """Remove undefined applications and computers."""
        log.info("Cleaning up applications and computers...")
        for appstatus in status.applications.copy():
            if not config.applications.find(appstatus.application):
                status.applications.remove(appstatus)
                log.info("Removed application: %s", appstatus)
            else:
                for computerstate in appstatus.computers.copy():
                    if not config.computers.find(computerstate.computer):
                        appstatus.computers.remove(computerstate)
                        log.info("Removed computer: %s", computerstate)

    @staticmethod
    def queue_all_applications(config, status, computer):
        """Queue applications for launch."""
        log.info("Queuing applications for launch...")
        for application in config.applications:
            if application.auto_queue:
                log.debug("Queuing %s on %s...", application, computer)
                status.queue(application, computer)

    @staticmethod
    def launch_queued_applications(config, status, computer, manager):
        """Launch applications that have been queued."""
        log.info("Launching queued applications...")
        for app_status in status.applications:
            if app_status.next:
                application = config.applications.get(app_status.application)
                show_queued(application, app_status.next)
                if app_status.next == computer:
                    latest = status.get_latest(application)
                    if latest in (computer, None) or application.no_wait:
                        if not manager.is_running(application):
                            manager.start(application)
                        app_status.next = None
                    else:
                        show_waiting(application, latest)
                elif manager.is_running(application):
                    manager.stop(application)

    @staticmethod
    def close_all_applications(config, manager):
        """Close all applications running on this computer."""
        log.info("Closing all applications on this computer...")
        for application in config.applications:
            manager.stop(application)

    @staticmethod
    def update_status(config, status, computer, manager):
        """Update each application's status."""
        log.info("Recording application status...")
        for application in config.applications:
            if manager.is_running(application):
                latest = status.get_latest(application)
                if computer != latest:
                    if status.is_running(application, computer):
                        # case 1: application just launched remotely
                        manager.stop(application)
                        status.stop(application, computer)
                        show_running(application, latest)
                        show_stopped(application, computer)
                    else:
                        # case 2: application just launched locally
                        status.start(application, computer)
                        show_running(application, computer)
                else:
                    # case 3: application already running locally
                    pass
            else:
                if status.is_running(application, computer):
                    # case 4: application just closed locally
                    status.stop(application, computer)
                    show_stopped(application, computer)
                else:
                    # case 5: application already closed locally
                    pass


def show_queued(application, computer):
    """Display the state of a queued application."""
    print("{} is queued for {}".format(application, computer))


def show_waiting(application, computer):
    """Display the old state of a running application."""
    print("{} is still running on {}".format(application, computer))


def show_running(application, computer):
    """Display the new state of a running application."""
    print("{} is now running on {}".format(application, computer))


def show_started(application, computer):
    """Display the new state of a started application."""
    print("{} is now started on {}".format(application, computer))


def show_stopped(application, computer):
    """Display the new state of a stopped application."""
    print("{} is now stopped on {}".format(application, computer))



"""Data structures for application/computer status."""

import functools
import logging

import yorm

from .timestamp import Timestamp

log = logging.getLogger(__name__)


def log_running(func):
    """Decorator for methods that return application status."""
    @functools.wraps(func)
    def wrapped(self, application, computer):
        """Wrapped method to log if an application is running."""
        running = func(self, application, computer)
        log.debug("%s marked as %s on: %s",
                  application, "started" if running else "stopped", computer)
        return running
    return wrapped


def log_starting(func):
    """Decorator for methods that mark an application as started."""
    @functools.wraps(func)
    def wrapped(self, application, computer):
        """Wrapped method to log that an application is started."""
        log.debug("Marking %s as started on %s...", application, computer)
        result = func(self, application, computer)
        log.debug("%s marked as started on: %s", application, computer)
        return result
    return wrapped


def log_stopping(func):
    """Decorator for methods that mark an application as stopped."""
    @functools.wraps(func)
    def wrapped(self, application, computer):
        """Wrapped method to log that an application is stopped."""
        log.debug("Marking %s as stopped on %s...", application, computer)
        result = func(self, application, computer)
        log.debug("%s marked as stopped on: %s", application, computer)
        return result
    return wrapped


@yorm.attr(computer=yorm.types.String)
@yorm.attr(timestamp=Timestamp)
class State(yorm.types.AttributeDictionary):
    """Dictionary of computer state."""

    def __init__(self, computer, timestamp=None):
        super().__init__()
        self.computer = computer
        self.timestamp = timestamp or Timestamp()

    def __str__(self):
        return str(self.computer)

    def __lt__(self, other):
        return str(self.computer).lower() < str(other.computer).lower()


@yorm.attr(all=State)
class StateList(yorm.types.SortedList):
    """List of computer states for an application."""


@yorm.attr(application=yorm.types.String)
@yorm.attr(computers=StateList)
@yorm.attr(next=yorm.types.NullableString)
class Status(yorm.types.AttributeDictionary):
    """Dictionary of computers using an application."""

    def __init__(self, application, computers=None, next=None):  # pylint: disable=redefined-builtin
        super().__init__()
        self.application = application
        self.computers = computers or StateList()
        self.next = next

    def __str__(self):
        return str(self.application)

    def __lt__(self, other):
        return str(self.application).lower() < str(other.application).lower()


@yorm.attr(all=Status)
class StatusList(yorm.types.SortedList):
    """List of application statuses."""


@yorm.attr(applications=StatusList)
@yorm.attr(counter=yorm.types.Integer)
class ProgramStatus(yorm.types.AttributeDictionary):
    """Dictionary of current program status."""

    def __init__(self, applications=None, counter=0):
        super().__init__()
        self.applications = applications or StatusList()
        self.counter = counter

    def find(self, application):
        """Return the application status for an application."""
        for app_status in self.applications:
            if app_status.application == application.name:
                break
        else:
            app_status = Status(application.name)
            self.applications.append(app_status)
        return app_status

    def get_latest(self, application):
        """Get the last computer's name logged as running an application."""
        for status in self.applications:
            if status.application == application.name:
                states = [s for s in status.computers if s.timestamp.active]
                if states:
                    states.sort(key=lambda s: s.timestamp, reverse=True)
                    log.debug("%s marked as started on: %s", application,
                              ', '.join(str(s) for s in states))
                    # TODO: consider returning the computer instance?
                    return states[0].computer

        log.debug("marked as started on: nothing")
        return None

    @log_running
    def is_running(self, application, computer):
        """Determine if an application is logged as running on a computer."""
        for status in self.applications:
            if status.application == application.name:
                for state in status.computers:
                    if state.computer == computer.name:
                        return state.timestamp.active

        # Status not found, assume the application is not running
        return False

    def queue(self, application, computer):
        """Record an application as queued for launch on a computer."""
        status = self.find(application)
        status.next = computer.name

    @log_starting
    def start(self, application, computer):
        """Record an application as running on a computer."""
        for status in self.applications:
            if status.application == application.name:
                for state in status.computers:
                    if state.computer == computer.name:
                        self.counter += 1
                        state.timestamp.started = self.counter
                        return
                break
        else:
            status = None

        # Status not found, add the application/computer as started
        self.counter += 1
        state = State(computer.name)
        state.timestamp.started = self.counter
        if status is None:
            status = Status(application.name)
            status.computers.append(state)
            self.applications.append(status)
        else:
            status.computers.append(state)

    @log_stopping
    def stop(self, application, computer):
        """Record an application as no longer running on a computer."""
        for status in self.applications:
            if status.application == application.name:
                for state in status.computers:
                    if state.computer == computer.name:
                        self.counter += 1
                        state.timestamp.stopped = self.counter
                        return
                break
        else:
            status = None

        # Status not found, add the application/computer as stopped
        self.counter += 1
        state = State(computer.name)
        state.timestamp.stopped = self.counter
        if status is None:
            status = Status(application.name)
            status.computers.append(state)
            self.applications.append(status)
        else:
            status.computers.append(state)



# pylint: disable=misplaced-comparison-constant,no-self-use

from mine.models import ProgramConfig


class TestProgramConfig:
    """Unit tests for the program configuration class."""

    def test_init(self):
        """Verify a new program configuration is blank."""
        config = ProgramConfig()
        assert not config.applications
        assert not config.computers



# pylint: disable=misplaced-comparison-constant,no-self-use,attribute-defined-outside-init

from unittest.mock import Mock

import pytest

from mine.models import State, Status, ProgramStatus, Application


class TestState:
    """Unit tests for the computer state class."""

    state1 = State('computer1')
    state2 = State('computer2')
    state2.timestamp.started = 42
    state3 = State('Computer2')
    state4 = State('Computer2')

    str_state = [
        ("computer1", state1),
        ("computer2", state2),
        ("Computer2", state3),
    ]

    @pytest.mark.parametrize("string,state", str_state)
    def test_str(self, string, state):
        """Verify states can be converted to strings."""
        assert string == str(state)

    def test_eq(self):
        """Verify states can be equated."""
        assert self.state3 == self.state4
        assert self.state2 != self.state3
        assert self.state1 != self.state2

    def test_lt(self):
        """Verify states can be sorted."""
        assert self.state1 < self.state2
        assert self.state3 > self.state1


class TestStatus:
    """Unit tests for the application status class."""

    status1 = Status('app1')
    status2 = Status('app2')
    status3 = Status('App2')
    status4 = Status('App2')

    str_status = [
        ("app1", status1),
        ("app2", status2),
        ("App2", status3),
    ]

    @pytest.mark.parametrize("string,status", str_status)
    def test_str(self, string, status):
        """Verify statuss can be converted to strings."""
        assert string == str(status)

    def test_eq(self):
        """Verify statuss can be equated."""
        assert self.status3 == self.status4
        assert self.status2 != self.status3
        assert self.status1 != self.status2

    def test_lt(self):
        """Verify statuss can be sorted."""
        assert self.status1 < self.status2
        assert self.status3 > self.status1


class TestProgramStatus:
    """Unit tests for the program status class."""

    def setup_method(self, _):
        """Create an empty program status for all tests."""
        self.status = ProgramStatus()
        self.application = Application('my-application')
        self.computer = Mock()
        self.computer.name = 'local'
        self.computer2 = Mock()
        self.computer2.name = 'remote'
        self.computer3 = Mock()
        self.computer3.name = 'remote2'

    def test_get_latest(self):
        """Verify the latest computer can be determined."""
        self.status.start(self.application, self.computer)
        self.status.start(self.application, self.computer2)
        self.status.stop(self.application, self.computer3)
        assert self.computer2.name == self.status.get_latest(self.application)
        assert 3 == self.status.counter

    def test_get_latest_empty(self):
        """Verify None is returned when there is no latest computer."""
        assert None is self.status.get_latest(self.application)
        assert 0 == self.status.counter

    def test_is_running_empty(self):
        """Verify no app is running when the list is empty."""
        assert False is self.status.is_running(self.application, self.computer)
        assert 0 == self.status.counter

    def test_queue(self):
        """Verify queuing an application sets the next computer."""
        self.status.queue(self.application, self.computer)
        app_status = self.status.find(self.application)
        assert self.computer.name == app_status.next

    def test_start(self):
        """Verify starting an application adds it to the list."""
        self.status.start(self.application, self.computer)
        names = [status.application for status in self.status.applications]
        assert self.application.name in names
        assert 1 == self.status.counter

    def test_stop(self):
        """Verify stopping an application adds it to the list."""
        self.status.stop(self.application, self.computer)
        names = [status.application for status in self.status.applications]
        assert self.application.name in names
        assert 1 == self.status.counter

    def test_start_again(self):
        """Verify starting a known application increments the counter."""
        self.status.start(self.application, self.computer)
        self.status.start(self.application, self.computer)
        self.status.start(self.application, self.computer2)
        assert self.status.is_running(self.application, self.computer)
        assert self.status.is_running(self.application, self.computer2)
        assert 3 == self.status.counter

    def test_stop_again(self):
        """Verify stopping a known application increments the counter."""
        self.status.stop(self.application, self.computer)
        self.status.stop(self.application, self.computer)
        self.status.stop(self.application, self.computer2)
        assert not self.status.is_running(self.application, self.computer)
        assert not self.status.is_running(self.application, self.computer2)
        assert 3 == self.status.counter



"""Shared base classes and mixins."""


class NameMixin:
    """Mixin class for objects identified by their name."""

    def __str__(self):
        return str(self.name)

    def __eq__(self, other):
        return str(self).lower() == str(other).lower()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return str(self).lower() < str(other).lower()



#!/usr/bin/env python

"""Command-line interface."""

import sys
import time
import argparse
import subprocess
import logging

import yorm

from . import CLI, VERSION, DESCRIPTION
from . import common
from . import services
from .manager import get_manager
from .models import Data, Application


log = logging.getLogger(__name__)
daemon = Application(CLI, filename=CLI)


def main(args=None):
    """Process command-line arguments and run the program."""

    # Shared options
    debug = argparse.ArgumentParser(add_help=False)
    debug.add_argument('-V', '--version', action='version', version=VERSION)
    group = debug.add_mutually_exclusive_group()
    group.add_argument('-v', '--verbose', action='count', default=0,
                       help="enable verbose logging")
    group.add_argument('-q', '--quiet', action='store_const', const=-1,
                       dest='verbose', help="only display errors and prompts")
    shared = {'formatter_class': common.HelpFormatter,
              'parents': [debug]}

    # Build main parser
    parser = argparse.ArgumentParser(prog=CLI, description=DESCRIPTION,
                                     **shared)
    parser.add_argument('-d', '--daemon', metavar='DELAY', nargs='?', const=300,
                        type=int, help="run continuously with delay [seconds]")
    parser.add_argument('-f', '--file', help="custom settings file path")
    subs = parser.add_subparsers(help="", dest='command', metavar="<command>")

    # Build switch parser
    info = "start applications on another computer"
    sub = subs.add_parser('switch', description=info.capitalize() + '.',
                          help=info, **shared)
    sub.add_argument('name', nargs='?',
                     help="computer to queue for launch (default: current)")

    # Build close parser
    info = "close applications on this computer"
    sub = subs.add_parser('close', description=info.capitalize() + '.',
                          help=info, **shared)

    # Build edit parser
    info = "launch the settings file for editing"
    sub = subs.add_parser('edit', description=info.capitalize() + '.',
                          help=info, **shared)

    # Build clean parser
    info = "display and delete conflicted files"
    sub = subs.add_parser('clean', description=info.capitalize() + '.',
                          help=info, **shared)
    sub.add_argument('-f', '--force', action='store_true',
                     help="actually delete the conflicted files")

    # Parse arguments
    args = parser.parse_args(args=args)
    kwargs = {'delay': args.daemon}
    if args.command == 'switch':
        kwargs['switch'] = args.name if args.name else True
    elif args.command == 'close':
        kwargs['switch'] = False
    elif args.command == 'edit':
        kwargs['edit'] = True
    elif args.command == 'clean':
        kwargs['delete'] = True
        kwargs['force'] = args.force

    # Configure logging
    common.configure_logging(args.verbose)

    # Run the program
    try:
        log.debug("Running main command...")
        success = run(path=args.file, **kwargs)
    except KeyboardInterrupt:
        msg = "command canceled"
        if common.verbosity == common.MAX_VERBOSITY:
            log.exception(msg)
        else:
            log.debug(msg)
        success = False
    if success:
        log.debug("Command succeeded")
    else:
        log.debug("Command failed")
        sys.exit(1)


def run(path=None, cleanup=True, delay=None,
        switch=None,
        edit=False,
        delete=False, force=False):
    """Run the program.

    :param path: custom settings file path
    :param cleanup: remove unused items from the config
    :param delay: number of seconds to delay before repeating

    :param switch: computer name to queue for launch

    :param edit: launch the configuration file for editing

    :param delete: attempt to delete conflicted files
    :param force: actually delete conflicted files

    """  # pylint: disable=too-many-arguments,too-many-branches
    manager = get_manager()
    if not manager.is_running(services.APPLICATION):
        manager.start(services.APPLICATION)

    root = services.find_root()
    path = path or services.find_config_path(root=root)

    data = Data()
    yorm.sync(data, path)

    config = data.config
    status = data.status

    log.info("Identifying current computer...")
    computer = config.computers.get_current()
    log.info("Current computer: %s", computer)

    if edit:
        return manager.launch(path)
    if delete:
        return services.delete_conflicts(root, force=force)
    if log.getEffectiveLevel() >= logging.WARNING:
        print("Updating application state...")

    if switch is True:
        switch = computer
    elif switch is False:
        data.close_all_applications(config, manager)
    elif switch:
        switch = config.computers.match(switch)

    if switch:
        if switch != computer:
            data.close_all_applications(config, manager)
        data.queue_all_applications(config, status, switch)

    while True:
        data.launch_queued_applications(config, status, computer, manager)
        data.update_status(config, status, computer, manager)

        if delay is None:
            break

        log.info("Delaying %s seconds for files to sync...", delay)
        time.sleep(delay)

        step = 1
        elapsed = 0
        log.info("Waiting %s seconds for status changes...", delay)
        while elapsed < delay and not data.modified:
            time.sleep(step)
            elapsed += step

        services.delete_conflicts(root, config_only=True, force=True)

    if cleanup:
        data.prune_status(config, status)

    if delay is None:
        return _restart_daemon(manager)

    return True


def _restart_daemon(manager):
    cmd = "nohup {} --daemon --verbose >> /tmp/mine.log 2>&1 &".format(CLI)
    if daemon and not manager.is_running(daemon):
        log.warning("Daemon is not running, attempting to restart...")

        log.info("$ %s", cmd)
        subprocess.call(cmd, shell=True)
        if manager.is_running(daemon):
            return True

        log.error("Manually start daemon: %s", cmd)
        return False

    return True


if __name__ == '__main__':  # pragma: no cover (manual test)
    main()
